from yucca.image_processing.transforms.YuccaTransform import YuccaTransform
import numpy as np
from typing import Union, Callable


# Stolen from Batchgenerators to avoid import error caused by deprecated modules imported in
# Batchgenerators.
def augment_gamma(
    data_sample,
    gamma_range=(0.5, 2),
    invert_image=False,
    epsilon=1e-7,
    per_channel=False,
    retain_stats: Union[bool, Callable[[], bool]] = False,
):
    if invert_image:
        data_sample = -data_sample

    if not per_channel:
        retain_stats_here = retain_stats() if callable(retain_stats) else retain_stats
        if retain_stats_here:
            mn = data_sample.mean()
            sd = data_sample.std()
        if np.random.random() < 0.5 and gamma_range[0] < 1:
            gamma = np.random.uniform(gamma_range[0], 1)
        else:
            gamma = np.random.uniform(max(gamma_range[0], 1), gamma_range[1])
        minm = data_sample.min()
        rnge = data_sample.max() - minm
        data_sample = np.power(((data_sample - minm) / float(rnge + epsilon)), gamma) * rnge + minm
        if retain_stats_here:
            data_sample = data_sample - data_sample.mean()
            data_sample = data_sample / (data_sample.std() + 1e-8) * sd
            data_sample = data_sample + mn
    else:
        for c in range(data_sample.shape[0]):
            retain_stats_here = retain_stats() if callable(retain_stats) else retain_stats
            if retain_stats_here:
                mn = data_sample[c].mean()
                sd = data_sample[c].std()
            if np.random.random() < 0.5 and gamma_range[0] < 1:
                gamma = np.random.uniform(gamma_range[0], 1)
            else:
                gamma = np.random.uniform(max(gamma_range[0], 1), gamma_range[1])
            minm = data_sample[c].min()
            rnge = data_sample[c].max() - minm
            data_sample[c] = np.power(((data_sample[c] - minm) / float(rnge + epsilon)), gamma) * float(rnge + epsilon) + minm
            if retain_stats_here:
                data_sample[c] = data_sample[c] - data_sample[c].mean()
                data_sample[c] = data_sample[c] / (data_sample[c].std() + 1e-8) * sd
                data_sample[c] = data_sample[c] + mn
    if invert_image:
        data_sample = -data_sample
    return data_sample


class Gamma(YuccaTransform):
    """
    WRAPPER FOR NNUNET AUGMENT GAMMA: https://github.com/MIC-DKFZ/batchgenerators/blob/8822a08a7dbfa4986db014e6a74b040778164ca6/batchgenerators/augmentations/color_augmentations.py

    Augments by changing 'gamma' of the image (same as gamma correction in photos or computer monitors

    :param gamma_range: range to sample gamma from. If one value is smaller than 1 and the other one is
    larger then half the samples will have gamma <1 and the other >1 (in the inverval that was specified).
    Tuple of float. If one value is < 1 and the other > 1 then half the images will be augmented with gamma values
    smaller than 1 and the other half with > 1
    :param invert_image: whether to invert the image before applying gamma augmentation
    :param retain_stats: Gamma transformation will alter the mean and std of the data in the patch. If retain_stats=True,
    the data will be transformed to match the mean and standard deviation before gamma augmentation. retain_stats
    can also be callable (signature retain_stats() -> bool)
    """

    def __init__(
        self,
        data_key="image",
        p_per_sample=1,
        p_invert_image=0.05,
        gamma_range=(0.5, 2.0),
        per_channel=True,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.gamma_range = gamma_range
        self.p_invert_image = p_invert_image
        self.per_channel = per_channel

    @staticmethod
    def get_params(p_invert_image):
        # No parameters to retrieve
        do_invert = False
        if np.random.uniform() < p_invert_image:
            do_invert = True
        return do_invert

    def __gamma__(self, imageVolume, gamma_range, invert_image, per_channel):
        return augment_gamma(imageVolume, gamma_range, invert_image, per_channel, retain_stats=False)

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            if np.random.uniform() < self.p_per_sample:
                do_invert = self.get_params(self.p_invert_image)
                data_dict[self.data_key][b] = self.__gamma__(
                    data_dict[self.data_key][b],
                    self.gamma_range,
                    do_invert,
                    per_channel=self.per_channel,
                )
        return data_dict



# From
# Copyright 2018 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module exposing surface distance based measures."""

from . import lookup_tables  # pylint: disable=relative-beyond-top-level
import numpy as np
from scipy import ndimage


def _assert_is_numpy_array(name, array):
    """Raises an exception if `array` is not a numpy array."""
    if not isinstance(array, np.ndarray):
        raise ValueError("The argument {!r} should be a numpy array, not a " "{}".format(name, type(array)))


def _check_nd_numpy_array(name, array, num_dims):
    """Raises an exception if `array` is not a `num_dims`-D numpy array."""
    if len(array.shape) != num_dims:
        raise ValueError("The argument {!r} should be a {}D array, not of " "shape {}".format(name, num_dims, array.shape))


def _check_2d_numpy_array(name, array):
    _check_nd_numpy_array(name, array, num_dims=2)


def _check_3d_numpy_array(name, array):
    _check_nd_numpy_array(name, array, num_dims=3)


def _assert_is_bool_numpy_array(name, array):
    _assert_is_numpy_array(name, array)
    if array.dtype != bool:
        raise ValueError("The argument {!r} should be a numpy array of type bool, " "not {}".format(name, array.dtype))


def _compute_bounding_box(mask):
    """Computes the bounding box of the masks.

    This function generalizes to arbitrary number of dimensions great or equal
    to 1.

    Args:
      mask: The 2D or 3D numpy mask, where '0' means background and non-zero means
        foreground.

    Returns:
      A tuple:
       - The coordinates of the first point of the bounding box (smallest on all
         axes), or `None` if the mask contains only zeros.
       - The coordinates of the second point of the bounding box (greatest on all
         axes), or `None` if the mask contains only zeros.
    """
    num_dims = len(mask.shape)
    bbox_min = np.zeros(num_dims, np.int64)
    bbox_max = np.zeros(num_dims, np.int64)

    # max projection to the x0-axis
    proj_0 = np.amax(mask, axis=tuple(range(num_dims))[1:])
    idx_nonzero_0 = np.nonzero(proj_0)[0]
    if len(idx_nonzero_0) == 0:  # pylint: disable=g-explicit-length-test
        return None, None

    bbox_min[0] = np.min(idx_nonzero_0)
    bbox_max[0] = np.max(idx_nonzero_0)

    # max projection to the i-th-axis for i in {1, ..., num_dims - 1}
    for axis in range(1, num_dims):
        max_over_axes = list(range(num_dims))  # Python 3 compatible
        max_over_axes.pop(axis)  # Remove the i-th dimension from the max
        max_over_axes = tuple(max_over_axes)  # numpy expects a tuple of ints
        proj = np.amax(mask, axis=max_over_axes)
        idx_nonzero = np.nonzero(proj)[0]
        bbox_min[axis] = np.min(idx_nonzero)
        bbox_max[axis] = np.max(idx_nonzero)

    return bbox_min, bbox_max


def _crop_to_bounding_box(mask, bbox_min, bbox_max):
    """Crops a 2D or 3D mask to the bounding box specified by `bbox_{min,max}`."""
    # we need to zeropad the cropped region with 1 voxel at the lower,
    # the right (and the back on 3D) sides. This is required to obtain the
    # "full" convolution result with the 2x2 (or 2x2x2 in 3D) kernel.
    # TODO:  This is correct only if the object is interior to the
    # bounding box.
    cropmask = np.zeros((bbox_max - bbox_min) + 2, np.uint8)

    num_dims = len(mask.shape)
    # pyformat: disable
    if num_dims == 2:
        cropmask[0:-1, 0:-1] = mask[bbox_min[0] : bbox_max[0] + 1, bbox_min[1] : bbox_max[1] + 1]
    elif num_dims == 3:
        cropmask[0:-1, 0:-1, 0:-1] = mask[
            bbox_min[0] : bbox_max[0] + 1, bbox_min[1] : bbox_max[1] + 1, bbox_min[2] : bbox_max[2] + 1
        ]
    # pyformat: enable
    else:
        assert False

    return cropmask


def _sort_distances_surfels(distances, surfel_areas):
    """Sorts the two list with respect to the tuple of (distance, surfel_area).

    Args:
      distances: The distances from A to B (e.g. `distances_gt_to_pred`).
      surfel_areas: The surfel areas for A (e.g. `surfel_areas_gt`).

    Returns:
      A tuple of the sorted (distances, surfel_areas).
    """
    sorted_surfels = np.array(sorted(zip(distances, surfel_areas)))
    return sorted_surfels[:, 0], sorted_surfels[:, 1]


def compute_surface_distances(mask_gt, mask_pred, spacing_mm):
    """Computes closest distances from all surface points to the other surface.

    This function can be applied to 2D or 3D tensors. For 2D, both masks must be
    2D and `spacing_mm` must be a 2-element list. For 3D, both masks must be 3D
    and `spacing_mm` must be a 3-element list. The description is done for the 2D
    case, and the formulation for the 3D case is present is parenthesis,
    introduced by "resp.".

    Finds all contour elements (resp surface elements "surfels" in 3D) in the
    ground truth mask `mask_gt` and the predicted mask `mask_pred`, computes their
    length in mm (resp. area in mm^2) and the distance to the closest point on the
    other contour (resp. surface). It returns two sorted lists of distances
    together with the corresponding contour lengths (resp. surfel areas). If one
    of the masks is empty, the corresponding lists are empty and all distances in
    the other list are `inf`.

    Args:
      mask_gt: 2-dim (resp. 3-dim) bool Numpy array. The ground truth mask.
      mask_pred: 2-dim (resp. 3-dim) bool Numpy array. The predicted mask.
      spacing_mm: 2-element (resp. 3-element) list-like structure. Voxel spacing
        in x0 anx x1 (resp. x0, x1 and x2) directions.

    Returns:
      A dict with:
      "distances_gt_to_pred": 1-dim numpy array of type float. The distances in mm
          from all ground truth surface elements to the predicted surface,
          sorted from smallest to largest.
      "distances_pred_to_gt": 1-dim numpy array of type float. The distances in mm
          from all predicted surface elements to the ground truth surface,
          sorted from smallest to largest.
      "surfel_areas_gt": 1-dim numpy array of type float. The length of the
        of the ground truth contours in mm (resp. the surface elements area in
        mm^2) in the same order as distances_gt_to_pred.
      "surfel_areas_pred": 1-dim numpy array of type float. The length of the
        of the predicted contours in mm (resp. the surface elements area in
        mm^2) in the same order as distances_gt_to_pred.

    Raises:
      ValueError: If the masks and the `spacing_mm` arguments are of incompatible
        shape or type. Or if the masks are not 2D or 3D.
    """
    # The terms used in this function are for the 3D case. In particular, surface
    # in 2D stands for contours in 3D. The surface elements in 3D correspond to
    # the line elements in 2D.

    _assert_is_bool_numpy_array("mask_gt", mask_gt)
    _assert_is_bool_numpy_array("mask_pred", mask_pred)

    if not len(mask_gt.shape) == len(mask_pred.shape) == len(spacing_mm):
        raise ValueError(
            "The arguments must be of compatible shape. Got mask_gt "
            "with {} dimensions ({}) and mask_pred with {} dimensions "
            "({}), while the spacing_mm was {} elements.".format(
                len(mask_gt.shape), mask_gt.shape, len(mask_pred.shape), mask_pred.shape, len(spacing_mm)
            )
        )

    num_dims = len(spacing_mm)
    if num_dims == 2:
        _check_2d_numpy_array("mask_gt", mask_gt)
        _check_2d_numpy_array("mask_pred", mask_pred)

        # compute the area for all 16 possible surface elements
        # (given a 2x2 neighbourhood) according to the spacing_mm
        neighbour_code_to_surface_area = lookup_tables.create_table_neighbour_code_to_contour_length(spacing_mm)
        kernel = lookup_tables.ENCODE_NEIGHBOURHOOD_2D_KERNEL
        full_true_neighbours = 0b1111
    elif num_dims == 3:
        _check_3d_numpy_array("mask_gt", mask_gt)
        _check_3d_numpy_array("mask_pred", mask_pred)

        # compute the area for all 256 possible surface elements
        # (given a 2x2x2 neighbourhood) according to the spacing_mm
        neighbour_code_to_surface_area = lookup_tables.create_table_neighbour_code_to_surface_area(spacing_mm)
        kernel = lookup_tables.ENCODE_NEIGHBOURHOOD_3D_KERNEL
        full_true_neighbours = 0b11111111
    else:
        raise ValueError("Only 2D and 3D masks are supported, not " "{}D.".format(num_dims))

    # compute the bounding box of the masks to trim the volume to the smallest
    # possible processing subvolume
    bbox_min, bbox_max = _compute_bounding_box(mask_gt | mask_pred)
    # Both the min/max bbox are None at the same time, so we only check one.
    if bbox_min is None:
        return {
            "distances_gt_to_pred": np.array([]),
            "distances_pred_to_gt": np.array([]),
            "surfel_areas_gt": np.array([]),
            "surfel_areas_pred": np.array([]),
        }

    # crop the processing subvolume.
    cropmask_gt = _crop_to_bounding_box(mask_gt, bbox_min, bbox_max)
    cropmask_pred = _crop_to_bounding_box(mask_pred, bbox_min, bbox_max)

    # compute the neighbour code (local binary pattern) for each voxel
    # the resulting arrays are spacially shifted by minus half a voxel in each
    # axis.
    # i.e. the points are located at the corners of the original voxels
    neighbour_code_map_gt = ndimage.filters.correlate(cropmask_gt.astype(np.uint8), kernel, mode="constant", cval=0)
    neighbour_code_map_pred = ndimage.filters.correlate(cropmask_pred.astype(np.uint8), kernel, mode="constant", cval=0)

    # create masks with the surface voxels
    borders_gt = (neighbour_code_map_gt != 0) & (neighbour_code_map_gt != full_true_neighbours)
    borders_pred = (neighbour_code_map_pred != 0) & (neighbour_code_map_pred != full_true_neighbours)

    # compute the distance transform (closest distance of each voxel to the
    # surface voxels)
    if borders_gt.any():
        distmap_gt = ndimage.morphology.distance_transform_edt(~borders_gt, sampling=spacing_mm)
    else:
        distmap_gt = np.Inf * np.ones(borders_gt.shape)

    if borders_pred.any():
        distmap_pred = ndimage.morphology.distance_transform_edt(~borders_pred, sampling=spacing_mm)
    else:
        distmap_pred = np.Inf * np.ones(borders_pred.shape)

    # compute the area of each surface element
    surface_area_map_gt = neighbour_code_to_surface_area[neighbour_code_map_gt]
    surface_area_map_pred = neighbour_code_to_surface_area[neighbour_code_map_pred]

    # create a list of all surface elements with distance and area
    distances_gt_to_pred = distmap_pred[borders_gt]
    distances_pred_to_gt = distmap_gt[borders_pred]
    surfel_areas_gt = surface_area_map_gt[borders_gt]
    surfel_areas_pred = surface_area_map_pred[borders_pred]

    # sort them by distance
    if distances_gt_to_pred.shape != (0,):
        distances_gt_to_pred, surfel_areas_gt = _sort_distances_surfels(distances_gt_to_pred, surfel_areas_gt)

    if distances_pred_to_gt.shape != (0,):
        distances_pred_to_gt, surfel_areas_pred = _sort_distances_surfels(distances_pred_to_gt, surfel_areas_pred)

    return {
        "distances_gt_to_pred": distances_gt_to_pred,
        "distances_pred_to_gt": distances_pred_to_gt,
        "surfel_areas_gt": surfel_areas_gt,
        "surfel_areas_pred": surfel_areas_pred,
    }


def compute_average_surface_distance(surface_distances):
    """Returns the average surface distance.

    Computes the average surface distances by correctly taking the area of each
    surface element into account. Call compute_surface_distances(...) before, to
    obtain the `surface_distances` dict.

    Args:
      surface_distances: dict with "distances_gt_to_pred", "distances_pred_to_gt"
      "surfel_areas_gt", "surfel_areas_pred" created by
      compute_surface_distances()

    Returns:
      A tuple with two float values:
        - the average distance (in mm) from the ground truth surface to the
          predicted surface
        - the average distance from the predicted surface to the ground truth
          surface.
    """
    distances_gt_to_pred = surface_distances["distances_gt_to_pred"]
    distances_pred_to_gt = surface_distances["distances_pred_to_gt"]
    surfel_areas_gt = surface_distances["surfel_areas_gt"]
    surfel_areas_pred = surface_distances["surfel_areas_pred"]
    average_distance_gt_to_pred = np.sum(distances_gt_to_pred * surfel_areas_gt) / np.sum(surfel_areas_gt)
    average_distance_pred_to_gt = np.sum(distances_pred_to_gt * surfel_areas_pred) / np.sum(surfel_areas_pred)
    return (average_distance_gt_to_pred, average_distance_pred_to_gt)


def compute_robust_hausdorff(surface_distances, percent):
    """Computes the robust Hausdorff distance.

    Computes the robust Hausdorff distance. "Robust", because it uses the
    `percent` percentile of the distances instead of the maximum distance. The
    percentage is computed by correctly taking the area of each surface element
    into account.

    Args:
      surface_distances: dict with "distances_gt_to_pred", "distances_pred_to_gt"
        "surfel_areas_gt", "surfel_areas_pred" created by
        compute_surface_distances()
      percent: a float value between 0 and 100.

    Returns:
      a float value. The robust Hausdorff distance in mm.
    """
    distances_gt_to_pred = surface_distances["distances_gt_to_pred"]
    distances_pred_to_gt = surface_distances["distances_pred_to_gt"]
    surfel_areas_gt = surface_distances["surfel_areas_gt"]
    surfel_areas_pred = surface_distances["surfel_areas_pred"]
    if len(distances_gt_to_pred) > 0:  # pylint: disable=g-explicit-length-test
        surfel_areas_cum_gt = np.cumsum(surfel_areas_gt) / np.sum(surfel_areas_gt)
        idx = np.searchsorted(surfel_areas_cum_gt, percent / 100.0)
        perc_distance_gt_to_pred = distances_gt_to_pred[min(idx, len(distances_gt_to_pred) - 1)]
    else:
        perc_distance_gt_to_pred = np.Inf

    if len(distances_pred_to_gt) > 0:  # pylint: disable=g-explicit-length-test
        surfel_areas_cum_pred = np.cumsum(surfel_areas_pred) / np.sum(surfel_areas_pred)
        idx = np.searchsorted(surfel_areas_cum_pred, percent / 100.0)
        perc_distance_pred_to_gt = distances_pred_to_gt[min(idx, len(distances_pred_to_gt) - 1)]
    else:
        perc_distance_pred_to_gt = np.Inf

    return max(perc_distance_gt_to_pred, perc_distance_pred_to_gt)


def compute_surface_overlap_at_tolerance(surface_distances, tolerance_mm):
    """Computes the overlap of the surfaces at a specified tolerance.

    Computes the overlap of the ground truth surface with the predicted surface
    and vice versa allowing a specified tolerance (maximum surface-to-surface
    distance that is regarded as overlapping). The overlapping fraction is
    computed by correctly taking the area of each surface element into account.

    Args:
      surface_distances: dict with "distances_gt_to_pred", "distances_pred_to_gt"
        "surfel_areas_gt", "surfel_areas_pred" created by
        compute_surface_distances()
      tolerance_mm: a float value. The tolerance in mm

    Returns:
      A tuple of two float values. The overlap fraction in [0.0, 1.0] of the
      ground truth surface with the predicted surface and vice versa.
    """
    distances_gt_to_pred = surface_distances["distances_gt_to_pred"]
    distances_pred_to_gt = surface_distances["distances_pred_to_gt"]
    surfel_areas_gt = surface_distances["surfel_areas_gt"]
    surfel_areas_pred = surface_distances["surfel_areas_pred"]
    rel_overlap_gt = np.sum(surfel_areas_gt[distances_gt_to_pred <= tolerance_mm]) / np.sum(surfel_areas_gt)
    rel_overlap_pred = np.sum(surfel_areas_pred[distances_pred_to_gt <= tolerance_mm]) / np.sum(surfel_areas_pred)
    return (rel_overlap_gt, rel_overlap_pred)


def compute_surface_dice_at_tolerance(surface_distances, tolerance_mm):
    """Computes the _surface_ DICE coefficient at a specified tolerance.

    Computes the _surface_ DICE coefficient at a specified tolerance. Not to be
    confused with the standard _volumetric_ DICE coefficient. The surface DICE
    measures the overlap of two surfaces instead of two volumes. A surface
    element is counted as overlapping (or touching), when the closest distance to
    the other surface is less or equal to the specified tolerance. The DICE
    coefficient is in the range between 0.0 (no overlap) to 1.0 (perfect overlap).

    Args:
      surface_distances: dict with "distances_gt_to_pred", "distances_pred_to_gt"
        "surfel_areas_gt", "surfel_areas_pred" created by
        compute_surface_distances()
      tolerance_mm: a float value. The tolerance in mm

    Returns:
      A float value. The surface DICE coefficient in [0.0, 1.0].
    """
    distances_gt_to_pred = surface_distances["distances_gt_to_pred"]
    distances_pred_to_gt = surface_distances["distances_pred_to_gt"]
    surfel_areas_gt = surface_distances["surfel_areas_gt"]
    surfel_areas_pred = surface_distances["surfel_areas_pred"]
    overlap_gt = np.sum(surfel_areas_gt[distances_gt_to_pred <= tolerance_mm])
    overlap_pred = np.sum(surfel_areas_pred[distances_pred_to_gt <= tolerance_mm])
    surface_dice = (overlap_gt + overlap_pred) / (np.sum(surfel_areas_gt) + np.sum(surfel_areas_pred))
    return surface_dice


def compute_dice_coefficient(mask_gt, mask_pred):
    """Computes soerensen-dice coefficient.

    compute the soerensen-dice coefficient between the ground truth mask `mask_gt`
    and the predicted mask `mask_pred`.

    Args:
      mask_gt: 3-dim Numpy array of type bool. The ground truth mask.
      mask_pred: 3-dim Numpy array of type bool. The predicted mask.

    Returns:
      the dice coeffcient as float. If both masks are empty, the result is NaN.
    """
    volume_sum = mask_gt.sum() + mask_pred.sum()
    if volume_sum == 0:
        return np.NaN
    volume_intersect = (mask_gt & mask_pred).sum()
    return 2 * volume_intersect / volume_sum



import nibabel as nib
import nibabel.processing as nibpro
from tqdm import tqdm
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subdirs
from yucca.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data
from sklearn.model_selection import train_test_split


def convert(path: str, subdir: str = "ISLES-2022"):
    # INPUT DATA
    path = f"{path}/{subdir}"
    file_suffix = ".nii.gz"

    # Train/Test Splits
    images_dir = join(path, "images")
    labels_dir = join(path, "labels_derivatives")

    training_samples, test_samples = train_test_split(subdirs(labels_dir, join=False), random_state=859032)

    ###OUTPUT DATA
    # Target names
    task_name = "Task005_ISLES22_FULL"
    prefix = "ISLES22_FULL"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    ###Populate Target Directory###
    # This is likely also the place to apply any re-orientation, resampling and/or label correction.
    for sTr in tqdm(training_samples, desc="Train"):
        label = nib.load(join(labels_dir, sTr, "ses-0001", sTr + "_ses-0001_msk" + file_suffix))

        dwi_file = nib.load(join(images_dir, sTr, "ses-0001", "dwi", sTr + "_ses-0001_dwi" + file_suffix))
        dwi_file = nibpro.resample_from_to(dwi_file, label, order=3)

        flair_file = nib.load(join(images_dir, sTr, "ses-0001", "anat", sTr + "_ses-0001_FLAIR" + file_suffix))
        flair_file = nibpro.resample_from_to(flair_file, label, order=3)

        adc_file = nib.load(join(images_dir, sTr, "ses-0001", "dwi", sTr + "_ses-0001_adc" + file_suffix))
        adc_file = nibpro.resample_from_to(adc_file, label, order=3)

        nib.save(flair_file, f"{target_imagesTr}/{prefix}_{sTr}_000.nii.gz")
        nib.save(dwi_file, f"{target_imagesTr}/{prefix}_{sTr}_001.nii.gz")
        nib.save(adc_file, f"{target_imagesTr}/{prefix}_{sTr}_002.nii.gz")
        nib.save(label, f"{target_labelsTr}/{prefix}_{sTr}.nii.gz")

    for sTs in tqdm(test_samples, desc="test"):
        label = nib.load(join(labels_dir, sTs, "ses-0001", sTs + "_ses-0001_msk" + file_suffix))

        dwi_file = nib.load(join(images_dir, sTs, "ses-0001", "dwi", sTs + "_ses-0001_dwi" + file_suffix))
        dwi_file = nibpro.resample_from_to(dwi_file, label, order=3)

        flair_file = nib.load(join(images_dir, sTs, "ses-0001", "anat", sTs + "_ses-0001_FLAIR" + file_suffix))
        flair_file = nibpro.resample_from_to(flair_file, label, order=3)

        adc_file = nib.load(join(images_dir, sTs, "ses-0001", "dwi", sTs + "_ses-0001_adc" + file_suffix))
        adc_file = nibpro.resample_from_to(adc_file, label, order=3)

        nib.save(flair_file, f"{target_imagesTs}/{prefix}_{sTs}_000.nii.gz")
        nib.save(dwi_file, f"{target_imagesTs}/{prefix}_{sTs}_001.nii.gz")
        nib.save(adc_file, f"{target_imagesTs}/{prefix}_{sTs}_002.nii.gz")
        nib.save(label, f"{target_labelsTs}/{prefix}_{sTs}.nii.gz")

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("Flair", "DWI", "ADC"),
        labels={0: "background", 1: "Infarct (ischemic stroke lesion)"},
        dataset_name=task_name,
        license="hands off!",
        dataset_description="ISLES-2022",
        dataset_reference="https://arxiv.org/abs/2206.06694",
    )



"""
This script can be used in two ways:
1. You can specify the paths to the folder containing predictions and labels/ground truth
using "--pred" and "--gt" and define the labels of interest using e.g. "-c 0 1"

2. You can specify the task, trainer and planner like it's also done in other yucca_ scripts.
if "-t" for target task is left blank, we assume you are predicting and evaluating data
from the same task as the one the model is trained on.
"""

import argparse
from batchgenerators.utilities.file_and_folder_operations import load_json, join
from yucca.evaluation.YuccaEvaluator import YuccaEvaluator
from yucca.utils.task_ids import maybe_get_task_from_task_id
from yucca.paths import yucca_raw_data, yucca_results


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-s",
        help="Name of the source task i.e. what the model is trained on. " "Should be of format: TaskXXX_MYTASK",
        required=False,
    )
    parser.add_argument(
        "-t",
        help="Name of the target task i.e. the data to be predicted. " "Should be of format: TaskXXX_MYTASK",
        required=False,
    )
    parser.add_argument("-m", help="Model Architecture. Defaults to UNet.", default="UNet")
    parser.add_argument("-d", help="2D, 25D or 3D model. Defaults to 3D.", default="3D")
    parser.add_argument(
        "-man",
        help="Full name of Manager Class. \n" "e.g. 'YuccaTrainer_DCE' or 'YuccaTrainer'. Defaults to YuccaTrainer.",
        default="YuccaManager",
    )
    parser.add_argument("-pl", help="Plan ID. Defaults to YuccaPlanner", default="YuccaPlanner")
    parser.add_argument("-chk", help="Checkpoint used for inference. Defaults to best.", default="best")
    parser.add_argument("-c", nargs="*", help="Classes to include for evaluation", type=str)

    # Alternatively, these can be used to manually specify paths
    parser.add_argument("--pred", help="manually specify path to predicted segmentations", default=None, required=False)
    parser.add_argument("--gt", help="manually specify path to ground truth", default=None, required=False)

    # Optionals (infrequently changed)
    parser.add_argument(
        "--task_type",
        default="segmentation",
        type=str,
        required=False,
        help="Defaults to segmentation. Set to 'classification' for classification tasks.",
    )
    parser.add_argument(
        "--as_binary", help="run evaluation as if data was binary", action="store_true", default=None, required=False
    )
    parser.add_argument(
        "--experiment",
        help="A name for the experiment being performed, with no spaces.",
        default="default",
    )
    parser.add_argument(
        "--no_wandb",
        help="Disable logging of evaluation results to wandb",
        default=False,
        action="store_true",
        required=False,
    )
    parser.add_argument("--obj_eval", help="enable object evaluation", action="store_true", default=None, required=False)
    parser.add_argument(
        "--overwrite",
        default=False,
        action="store_true",
        required=False,
        help="Overwrite existing predictions",
    )
    parser.add_argument("--split_idx", type=int, help="idx of splits to use for training.", default=0)
    parser.add_argument(
        "--split_data_method", help="Specify splitting method. Either kfold, simple_train_val_split", default="kfold"
    )
    parser.add_argument(
        "--split_data_param",
        help="Specify the parameter for the selected split method. For KFold use an int, for simple_split use a float between 0.0-1.0.",
        default=5,
    )
    parser.add_argument("--surface_eval", help="enable surface evaluation", action="store_true", default=None, required=False)

    parser.add_argument(
        "--version", "-v", help="version number of the model. Defaults to 0.", default=0, type=int, required=False
    )
    args = parser.parse_args()

    source_task = maybe_get_task_from_task_id(args.s)
    target_task = maybe_get_task_from_task_id(args.t)
    manager_name = args.man
    model = args.m
    dimensions = args.d
    plan_id = args.pl
    checkpoint = args.chk
    obj = args.obj_eval
    overwrite = args.overwrite
    as_binary = args.as_binary
    classes = args.c
    predpath = args.pred
    gtpath = args.gt
    num_version = args.version
    task_type = args.task_type
    use_wandb = not args.no_wandb
    split_idx = args.split_idx
    split_data_method = args.split_data_method
    split_data_param = args.split_data_param
    surface_eval = args.surface_eval
    assert (predpath and gtpath) or source_task, "Either supply BOTH paths or the source task"

    if not predpath:
        if not target_task:
            target_task = source_task

        predpath = join(  # TODO: Extract this into a function
            yucca_results,
            target_task,
            source_task,
            model + "__" + dimensions,
            manager_name + "__" + plan_id,
            f"{split_data_method}_{split_data_param}_fold_{split_idx}",
            "version_" + str(num_version),
            checkpoint,
        )
        gtpath = join(yucca_raw_data, target_task, "labelsTs")
        classes = list(load_json(join(yucca_raw_data, target_task, "dataset.json"))["labels"].keys())

    evaluator = YuccaEvaluator(
        classes,
        folder_with_predictions=predpath,
        folder_with_ground_truth=gtpath,
        do_object_eval=obj,
        do_surface_eval=surface_eval,
        overwrite=overwrite,
        as_binary=as_binary,
        task_type=task_type,
        use_wandb=use_wandb,
    )
    evaluator.run()


if __name__ == "__main__":
    main()



import gzip
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.task_conversion.utils import generate_dataset_json, dirs_in_dir, should_use_volume
from yucca.paths import yucca_raw_data
from tqdm import tqdm
import shutil

import nibabel as nib


def convert(path: str, subdir: str = "ADNI_NIFTI"):
    """INPUT DATA - Define input path and suffixes"""

    print("Converting task207 with try except")

    path = join(path, subdir)
    ext = ".nii"

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task207_ADNI"
    task_prefix = "ADNI"

    """ Access the input data. If images are not split into train/test, and you wish to randomly
    split the data, uncomment and adapt the following lines to fit your local path. """

    subjects_dir = join(path, "ADNI")
    target_base = join(yucca_raw_data, task_name)
    target_imagesTr = join(target_base, "imagesTr")

    maybe_mkdir_p(target_imagesTr)

    """Populate Target Directory
    This is also the place to apply any re-orientation, resampling and/or label correction."""

    skipped = []
    errors = []

    for subject in tqdm(dirs_in_dir(subjects_dir), desc="Subjects"):
        subject_dir = join(subjects_dir, subject)
        for modality in dirs_in_dir(subject_dir):
            modality_dir = join(subject_dir, modality)
            for session in dirs_in_dir(modality_dir):
                session_dir = join(modality_dir, session)
                for id in dirs_in_dir(session_dir):
                    id_dir = join(session_dir, id)
                    for file in subfiles(id_dir, join=False, suffix=ext):
                        image_path = join(id_dir, file)
                        file_name = file[: -len(ext)]
                        try:
                            vol = nib.load(image_path)
                            if should_use_volume(vol):
                                output_name = f"{task_prefix}_{file_name}_000.nii.gz"
                                output_path = join(target_imagesTr, output_name)
                                image_file = open(image_path, "rb")
                                shutil.copyfileobj(image_file, gzip.open(output_path, "wb"))
                            else:
                                skipped.append(image_path)
                        except:
                            errors.append(image_path)

    print("skipped volumes:", skipped)
    print("erroneous volumes", errors)

    print("num errors", len(errors), "num skipped", len(skipped))

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        imagesTs_dir=None,
        modalities=["MRI"],
        labels=None,
        dataset_name=task_name,
        license="you should check this if you dont know",
        dataset_description="ADNI",
        dataset_reference="ADNI",
    )



import torch
from torchmetrics.classification import ConfusionMatrix


def torch_confusion_matrix_from_logits(pred, seg):
    if len(seg.shape) == len(pred.shape):
        assert seg.shape[1] == 1
        seg = seg[:, 0]

    n_classes = pred.shape[1]

    cfm = ConfusionMatrix(task="multiclass", num_classes=n_classes)

    if torch.cuda.is_available():
        cfm = cfm.to("cuda")
    matrix = cfm(pred, seg)
    return matrix


def torch_get_tp_fp_tn_fn(confusion_matrix, ignore_label=0):
    TP = []
    FP = []
    TN = []
    FN = []

    for label in range(confusion_matrix.shape[0]):
        if label == ignore_label:
            continue
        tp = confusion_matrix[label, label]
        fp = torch.sum(confusion_matrix[:, label]) - tp
        fn = torch.sum(confusion_matrix[label]) - tp
        tn = torch.sum(confusion_matrix) - tp - fp - fn
        TP.append(tp.cpu().numpy())
        FP.append(fp.cpu().numpy())
        TN.append(tn.cpu().numpy())
        FN.append(fn.cpu().numpy())
    return TP, FP, TN, FN



from yucca.image_processing.transforms.YuccaTransform import YuccaTransform
import numpy as np
from typing import Tuple


class AdditiveNoise(YuccaTransform):
    def __init__(self, data_key="image", p_per_sample=1, mean=(0.0, 0.0), sigma=(1e-3, 1e-4)):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.mean = mean
        self.sigma = sigma

    @staticmethod
    def get_params(mean: Tuple[float], sigma: Tuple[float]) -> Tuple[float]:
        mean = float(np.random.uniform(*mean))
        sigma = float(np.random.uniform(*sigma))
        return mean, sigma

    def __additiveNoise__(self, imageVolume, mean, sigma):
        # J = I+n
        gauss = np.random.normal(mean, sigma, imageVolume.shape)
        return imageVolume + gauss

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (c, x, y, z) or (c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                mean, sigma = self.get_params(self.mean, self.sigma)
                if np.random.uniform() < self.p_per_sample:
                    data_dict[self.data_key][b, c] = self.__additiveNoise__(data_dict[self.data_key][b, c], mean, sigma)
        return data_dict


class MultiplicativeNoise(YuccaTransform):
    """
    variables in DIKU_3D_augmentation_params:
        do_multiplicativeNoise
        multiplicativeNoise_p_per_sample
        multiplicativeNoise_mean
        multiplicativeNoise_sigma
    """

    def __init__(self, data_key="image", p_per_sample=1, mean=(0.0, 0.0), sigma=(1e-3, 1e-4)):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.mean = mean
        self.sigma = sigma

    @staticmethod
    def get_params(mean: Tuple[float], sigma: Tuple[float]) -> Tuple[float]:
        mean = float(np.random.uniform(*mean))
        sigma = float(np.random.uniform(*sigma))
        return mean, sigma

    def __multiplicativeNoise__(self, imageVolume, mean, sigma):
        # J = I + I*n
        gauss = np.random.normal(mean, sigma, imageVolume.shape)
        return imageVolume + imageVolume * gauss

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                if np.random.uniform() < self.p_per_sample:
                    mean, sigma = self.get_params(self.mean, self.sigma)
                    data_dict[self.data_key][b, c] = self.__multiplicativeNoise__(data_dict[self.data_key][b, c], mean, sigma)
        return data_dict



from dataclasses import dataclass
from typing import Tuple, Union, Optional, Literal, Iterable
from yucca.network_architectures.utils.model_memory_estimation import find_optimal_tensor_dims
import logging


@dataclass
class InputDimensionsConfig:
    batch_size: int
    patch_size: Union[Tuple[int, int], Tuple[int, int, int]]
    num_modalities: int

    def lm_hparams(self):
        return {
            "batch_size": self.batch_size,
            "patch_size": self.patch_size,
            "num_modalities": self.num_modalities,
        }


def get_input_dims_config(
    plan: dict,
    model_dimensions: Literal["2D", "3D"],
    model_name: str,
    num_classes: int,
    ckpt_patch_size: Optional[Iterable[int]] = None,
    max_vram: Optional[int] = None,
    batch_size: Union[int, Literal["tiny"]] = None,
    patch_based_training: bool = True,
    patch_size: Union[Literal["max", "min", "mean", "tiny"], Tuple[int, int], Tuple[int, int, int], None] = None,
):
    """
    The priority list for setting patch_size:
    A.
        1. Patch based training is false = we train on the full image size of the dataset.
        - This requires the dataset to be preprocessed with a planner that uses "fixed_target_size" such as
        the /planning/resampling/YuccaPlanner_MaxSize.py
    B.
        1. Patch size is set in the manager. This overrides other defaults and inferred patch sizes.
        2. Patch size is set in the ckpt.
        - Relevant for finetuning and inference. This happens if patch size was originally set using the manager.
        In this case we have patch size X from the manager and "default" patch size Y in the plans, where X
        originally overruled Y and thus should do so again.
        3. Patch size is inferred in the plans.
    """
    num_modalities = max(1, plan.get("num_modalities") or len(plan["dataset_properties"]["modalities"]))

    if isinstance(batch_size, str):
        if batch_size == "tiny":
            batch_size = 2
        else:
            raise ValueError(f"Unknown batch size param: {batch_size}")

    # A.1. Get patch size from full image size if we are not doing patch_based_training.
    if not patch_based_training:
        assert plan.get("new_max_size") == plan.get(
            "new_min_size"
        ), "sizes in dataset are not uniform. Non-patch based training only works for datasets with uniform data shapes."
        patch_size = tuple(plan.get("new_max_size"))
        logging.info(f"Getting patch size for non-patch based training")
    else:
        # B.1. Try get patch from manager
        if patch_size is not None:
            logging.info(f"Getting patch size based on manual input of: {patch_size}")
            # Can be three things here: 1. a list/tuple of ints, 2. a list of one int/str or 3. just an int/str
            # First check case 1.
            if isinstance(patch_size, (list, tuple)) and len(patch_size) > 1:
                patch_size = tuple(int(n) for n in patch_size)
            else:
                # Then check case 2 and convert to be identical to case 3.
                if isinstance(patch_size, list) and len(patch_size) == 1:
                    patch_size = patch_size[0]
                # Proceed as if case 3.
                if patch_size in ["max", "min", "mean"]:
                    patch_size = tuple(plan[f"new_{patch_size}_size"])
                elif patch_size == "tiny":
                    patch_size = (32, 32, 32)
                elif patch_size.isdigit():
                    patch_size = (int(patch_size),) * 3
                else:
                    raise ValueError(f"Unknown patch size param: {patch_size}")

        # B.2. Try get patch from ckpt
        elif ckpt_patch_size is not None:
            patch_size = ckpt_patch_size
            logging.info(f"Using patch size found in checkpoint: {ckpt_patch_size}")

        # B.3. Infer patch size from constraints
        else:
            logging.info("Patch size will be infered from hardware constraints")

    # If we have now selected a 3D patch for a 2D model we skip the first dim
    # as we will be extracting patches from that dimension.
    # This also includes if we are not doing patch_based_training.
    if patch_size is not None and model_dimensions == "2D" and len(patch_size) == 3:
        patch_size = patch_size[1:]

    # We ALWAYS run this because even in the case that we have patch_size AND batch_size
    # this function will make sure they're valid by e.g. making the patch size divisible by 16.
    batch_size, patch_size = find_optimal_tensor_dims(
        fixed_patch_size=patch_size,
        fixed_batch_size=batch_size,
        dimensionality=model_dimensions,
        num_classes=num_classes,
        modalities=num_modalities,
        model_name=model_name,
        max_patch_size=plan["new_mean_size"],
        max_memory_usage_in_gb=max_vram,
    )

    assert isinstance(patch_size, tuple), patch_size
    assert isinstance(batch_size, int), batch_size
    assert batch_size > 0, batch_size
    assert (model_dimensions == "2D" and len(patch_size) == 2) or (model_dimensions == "3D" and len(patch_size) == 3), (
        model_dimensions,
        len(patch_size),
    )

    if plan.get("patch_size") and plan.get("patch_size") != patch_size:
        raise UserWarning(f"Overwriting patch size from plan ({plan.get('patch_size')}) with {patch_size}")
    if plan.get("batch_size") and plan.get("batch_size") != batch_size:
        raise UserWarning(f"Overwriting batch size from plan ({plan.get('batch_size')}) with {batch_size}")

    logging.info(f"Using batch size: {batch_size} and patch size: {patch_size}")

    return InputDimensionsConfig(
        batch_size=batch_size,
        patch_size=patch_size,
        num_modalities=num_modalities,
    )



import lightning as pl
import torchvision
import logging
import torch
from typing import Literal, Optional, Union
from torch.utils.data import DataLoader, Sampler
from batchgenerators.utilities.file_and_folder_operations import join
from yucca.training.configuration.configure_input_dims import InputDimensionsConfig
from yucca.training.configuration.split_data import SplitConfig
from yucca.training.data_loading.YuccaDataset import YuccaTestDataset, YuccaTrainDataset
from yucca.training.data_loading.samplers import InfiniteRandomSampler


class YuccaDataModule(pl.LightningDataModule):
    """
    The YuccaDataModule class is a PyTorch Lightning DataModule designed for handling data loading
    and preprocessing in the context of the Yucca project.

    It extends the pl.LightningDataModule class and provides methods for preparing data, setting up
    datasets for training, validation, and prediction, as well as creating data loaders for these stages.

    configurator (YuccaConfigurator): An instance of the YuccaConfigurator class containing configuration parameters.

    composed_train_transforms (torchvision.transforms.Compose, optional): A composition of transforms to be applied to the training dataset. Default is None.
    composed_val_transforms (torchvision.transforms.Compose, optional): A composition of transforms to be applied to the validation dataset. Default is None.

    num_workers (int, optional): Number of workers for data loading. Default is 8.

    pred_data_dir (str, optional): Directory containing data for prediction. Required only during the "predict" stage.

    pre_aug_patch_size (list or tuple, optional): Patch size before data augmentation. Default is None.
        - The purpose of the pre_aug_patch_size is to increase computational efficiency while not losing important information.
        If we have a volume of 512x512x512 and our model only works with patches of 128x128x128 there's no reason to
        apply the transform to the full volume. To avoid this we crop the volume before transforming it.

        But, we do not want to crop it to 128x128x128 before transforming it. Especially not before applying spatial transforms.
        Both because
        (1) the edges will contain a lot of border interpolation artifacts, and
        (2) if we crop to 128x128x128 and then rotate the image 45 degrees or downscale it (zoom out effect)
        we suddenly introduce dark areas where they should not be. We could've simply kept more of the original
        volume BEFORE scaling or rotating, then our 128x128x128 wouldn't be part-black.

        Therefore the pre_aug_patch_size parameter allows users to specify a patch size before augmentation is applied.
        This can potentially avoid dark or low-intensity areas at the borders and it also helps mitigate the risk of
        introducing artifacts during data augmentation, especially in regions where interpolation may have a significant impact.
    """

    def __init__(
        self,
        input_dims_config: InputDimensionsConfig,
        image_extension: str,
        composed_train_transforms: Optional[torchvision.transforms.Compose] = None,
        composed_val_transforms: Optional[torchvision.transforms.Compose] = None,
        num_workers: Optional[int] = None,
        overwrite_predictions: bool = False,
        pred_data_dir: Optional[str] = None,
        pred_save_dir: Optional[str] = None,
        pre_aug_patch_size: Optional[Union[list, tuple]] = None,
        splits_config: Optional[SplitConfig] = None,
        split_idx: Optional[int] = None,
        task_type: Optional[str] = None,
        train_sampler: Optional[Sampler] = InfiniteRandomSampler,
        val_sampler: Optional[Sampler] = InfiniteRandomSampler,
        train_data_dir: Optional[str] = None,
    ):
        super().__init__()

        # extract parameters
        self.batch_size = input_dims_config.batch_size
        self.patch_size = input_dims_config.patch_size
        self.image_extension = image_extension
        self.task_type = task_type

        self.split_idx = split_idx
        self.splits_config = splits_config
        self.train_data_dir = train_data_dir

        # Set by initialize()
        self.composed_train_transforms = composed_train_transforms
        self.composed_val_transforms = composed_val_transforms
        self.pre_aug_patch_size = pre_aug_patch_size

        # Set in the predict loop
        self.overwrite_predictions = overwrite_predictions
        self.pred_data_dir = pred_data_dir
        self.pred_save_dir = pred_save_dir

        # Set default values
        self.num_workers = max(0, int(torch.get_num_threads() - 1)) if num_workers is None else num_workers
        self.val_num_workers = self.num_workers
        self.train_sampler = train_sampler
        self.val_sampler = val_sampler
        logging.info(f"Using {self.num_workers} workers")

    def setup(self, stage: Literal["fit", "test", "predict"]):
        logging.info(f"Setting up data for stage: {stage}")
        expected_stages = ["fit", "test", "predict"]
        assert stage in expected_stages, "unexpected stage. " f"Expected: {expected_stages} and found: {stage}"

        # Assign train/val datasets for use in dataloaders
        if stage == "fit":
            assert self.train_data_dir is not None
            assert self.split_idx is not None
            assert self.splits_config is not None
            assert self.task_type is not None

            self.train_samples = [join(self.train_data_dir, i) for i in self.splits_config.train(self.split_idx)]
            self.val_samples = [join(self.train_data_dir, i) for i in self.splits_config.val(self.split_idx)]

            if len(self.train_samples) < 100:
                logging.info(f"Training on samples: {self.train_samples}")

            if len(self.val_samples) < 100:
                logging.info(f"Validating on samples: {self.val_samples}")

            self.train_dataset = YuccaTrainDataset(
                self.train_samples,
                composed_transforms=self.composed_train_transforms,
                patch_size=self.pre_aug_patch_size if self.pre_aug_patch_size is not None else self.patch_size,
                task_type=self.task_type,
            )

            self.val_dataset = YuccaTrainDataset(
                self.val_samples,
                composed_transforms=self.composed_val_transforms,
                patch_size=self.patch_size,
                task_type=self.task_type,
            )

        if stage == "predict":
            assert self.pred_data_dir is not None, "set a pred_data_dir for inference to work"
            # This dataset contains ONLY the images (and not the labels)
            # It will return a tuple of (case, case_id)
            self.pred_dataset = YuccaTestDataset(
                self.pred_data_dir,
                pred_save_dir=self.pred_save_dir,
                overwrite_predictions=self.overwrite_predictions,
                suffix=self.image_extension,
            )

    def train_dataloader(self):
        logging.info(f"Starting training with data from: {self.train_data_dir}")
        sampler = self.train_sampler(self.train_dataset) if self.train_sampler is not None else None
        return DataLoader(
            self.train_dataset,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
            pin_memory=torch.cuda.is_available(),
            sampler=sampler,
            shuffle=sampler is None,
        )

    def val_dataloader(self):
        sampler = self.val_sampler(self.val_dataset) if self.val_sampler is not None else None
        return DataLoader(
            self.val_dataset,
            num_workers=self.val_num_workers,
            batch_size=self.batch_size,
            pin_memory=torch.cuda.is_available(),
            sampler=sampler,
        )

    def test_dataloader(self):
        return None

    def predict_dataloader(self):
        logging.info("Starting inference")
        return DataLoader(self.pred_dataset, num_workers=self.num_workers, batch_size=1)



from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.task_conversion.utils import generate_dataset_json, dirs_in_dir
from yucca.paths import yucca_raw_data
from tqdm import tqdm
import shutil


def convert(path: str, subdir: str = "ISLES-2022"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)
    ext = ".nii.gz"

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task202_ISLES22"
    task_prefix = "ISLES22"

    """ Access the input data. If images are not split into train/test, and you wish to randomly
    split the data, uncomment and adapt the following lines to fit your local path. """

    subjects_dir = join(path, "images")
    target_base = join(yucca_raw_data, task_name)
    target_imagesTr = join(target_base, "imagesTr")

    maybe_mkdir_p(target_imagesTr)

    """Populate Target Directory
    This is also the place to apply any re-orientation, resampling and/or label correction."""

    for subject in tqdm(dirs_in_dir(subjects_dir), desc="Subjects"):
        subject_dir = join(subjects_dir, subject)
        for session in dirs_in_dir(subject_dir):
            session_dir = join(subject_dir, session)
            for modality in dirs_in_dir(session_dir):
                modality_dir = join(session_dir, modality)
                for file in subfiles(modality_dir, join=False, suffix=ext):
                    image_path = join(modality_dir, file)
                    file_name = file[: -len(ext)]
                    output_name = f"{task_prefix}_{file_name}_000.nii.gz"
                    output_path = join(target_imagesTr, output_name)
                    shutil.copy2(image_path, output_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        imagesTs_dir=None,
        modalities=["MRI"],
        labels=None,
        dataset_name=task_name,
        license="CC-BY-SA 4.0",
        dataset_description="Decathlon: Brain Tumour Segmentation",
        dataset_reference="King's College London",
    )



"""
Takes raw data conforming with Yucca standards and preprocesses according to the generic scheme
"""

import numpy as np
import re
import os
import logging
import time
from yucca.preprocessing.YuccaPreprocessor import YuccaPreprocessor
from yucca.paths import yucca_preprocessed_data, yucca_raw_data
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    subfiles,
    save_pickle,
    isfile,
)


class UnsupervisedPreprocessor(YuccaPreprocessor):
    def initialize_paths(self):
        # Have to overwrite how we get the subject_ids as there's no labelsTr to get them from.
        # Therefore we use the imagesTr folder and remove the modality suffix.
        self.target_dir = join(yucca_preprocessed_data, self.task, self.plans["plans_name"])
        self.input_dir = join(yucca_raw_data, self.task)
        self.imagepaths = subfiles(join(self.input_dir, "imagesTr"), suffix=self.image_extension)

        subject_ids = subfiles(join(self.input_dir, "imagesTr"), suffix=self.image_extension, join=False)
        self.subject_ids = [re.sub(r"_\d+\.", ".", subject) for subject in subject_ids]

    def preprocess_train_subject(self, subject_id):
        subject_id = subject_id.split(os.extsep, 1)[0]
        arraypath = join(self.target_dir, subject_id + ".npy")
        picklepath = join(self.target_dir, subject_id + ".pkl")

        if isfile(arraypath) and isfile(picklepath):
            logging.info(f"Case: {subject_id} already exists. Skipping.")
            return

        start_time = time.time()

        images, _, image_props = self._preprocess_train_subject(subject_id, label_exists=False, preprocess_label=False)

        images = np.array(images)

        # save the image
        np.save(arraypath, images)

        # save metadata as .pkl
        save_pickle(image_props, picklepath)

        end_time = time.time()
        logging.info(
            f"Preprocessed case: {subject_id} \n"
            f"size before: {image_props['original_size']} size after: {image_props['new_size']} \n"
            f"spacing before: {image_props['original_spacing']} spacing after: {image_props['new_spacing']} \n"
            f"Saving {subject_id} in {arraypath} \n"
            f"Time elapsed: {round(end_time-start_time, 4)} \n"
        )



"""
https://www.meccanismocomplesso.org/en/3d-rotations-and-euler-angles-in-python/
From: https://stackoverflow.com/questions/59738230/apply-rotation-defined-by-euler-angles-to-3d-image-in-python
"""

import numpy as np
from scipy.ndimage import map_coordinates
from yucca.image_processing.matrix_ops import (
    create_zero_centered_coordinate_matrix,
    deform_coordinate_matrix,
    Rx,
    Ry,
    Rz,
    Rz2D,
)
from yucca.image_processing.transforms.YuccaTransform import YuccaTransform
from typing import Tuple


class Spatial(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        label_key="label",
        crop=False,
        cval="min",
        patch_size: Tuple[int] = None,
        random_crop=True,
        p_deform_per_sample=1,
        deform_sigma=(20, 30),
        deform_alpha=(300, 600),
        p_rot_per_sample=1,
        p_rot_per_axis=1,
        x_rot_in_degrees=(0.0, 10.0),
        y_rot_in_degrees=(0.0, 10.0),
        z_rot_in_degrees=(0.0, 10.0),
        p_scale_per_sample=1,
        scale_factor=(0.85, 1.15),
        skip_label=False,
    ):
        self.data_key = data_key
        self.label_key = label_key
        self.skip_label = skip_label
        self.do_crop = crop
        self.cval = cval
        self.patch_size = patch_size
        self.random_crop = random_crop

        self.p_deform_per_sample = p_deform_per_sample
        self.deform_sigma = deform_sigma
        self.deform_alpha = deform_alpha

        self.p_rot_per_sample = p_rot_per_sample
        self.p_rot_per_axis = p_rot_per_axis
        self.x_rot_in_degrees = x_rot_in_degrees
        self.y_rot_in_degrees = y_rot_in_degrees
        self.z_rot_in_degrees = z_rot_in_degrees

        self.p_scale_per_sample = p_scale_per_sample
        self.scale_factor = scale_factor

    @staticmethod
    def get_params(
        deform_alpha: Tuple[float],
        deform_sigma: Tuple[float],
        x_rot: Tuple[float],
        y_rot: Tuple[float],
        z_rot: Tuple[float],
        scale_factor: Tuple[float],
    ) -> Tuple[float]:
        if deform_alpha:
            deform_alpha = float(np.random.uniform(*deform_alpha))
        if deform_sigma:
            deform_sigma = float(np.random.uniform(*deform_sigma))

        if x_rot:
            x_rot = float(np.random.uniform(*x_rot)) * (np.pi / 180)
        if y_rot:
            y_rot = float(np.random.uniform(*y_rot)) * (np.pi / 180)
        if z_rot:
            z_rot = float(np.random.uniform(*z_rot)) * (np.pi / 180)

        if scale_factor:
            scale_factor = float(np.random.uniform(*scale_factor))

        return deform_alpha, deform_sigma, x_rot, y_rot, z_rot, scale_factor

    def __CropDeformRotateScale__(
        self,
        data_dict,
        patch_size,
        alpha,
        sigma,
        x_rot,
        y_rot,
        z_rot,
        scale_factor,
        skip_label,
    ):
        image = data_dict[self.data_key]
        if not self.do_crop:
            patch_size = image.shape[2:]
        if self.cval == "min":
            cval = float(image.min())
        else:
            cval = self.cval
        assert isinstance(cval, (int, float)), f"got {cval} of type {type(cval)}"

        coords = create_zero_centered_coordinate_matrix(patch_size)
        imageCanvas = np.zeros((image.shape[0], image.shape[1], *patch_size), dtype=np.float32)

        # First we apply deformation to the coordinate matrix
        if np.random.uniform() < self.p_deform_per_sample:
            coords = deform_coordinate_matrix(coords, alpha=alpha, sigma=sigma)

        # Then we rotate the coordinate matrix around one or more axes
        if np.random.uniform() < self.p_rot_per_sample:
            rot_matrix = np.eye(len(patch_size))
            if len(patch_size) == 2:
                rot_matrix = np.dot(rot_matrix, Rz2D(z_rot))
            else:
                if np.random.uniform() < self.p_rot_per_axis:
                    rot_matrix = np.dot(rot_matrix, Rx(x_rot))
                if np.random.uniform() < self.p_rot_per_axis:
                    rot_matrix = np.dot(rot_matrix, Ry(y_rot))
                if np.random.uniform() < self.p_rot_per_axis:
                    rot_matrix = np.dot(rot_matrix, Rz(z_rot))

            coords = np.dot(coords.reshape(len(patch_size), -1).transpose(), rot_matrix).transpose().reshape(coords.shape)

        # And finally scale it
        # Scaling effect is "inverted"
        # i.e. a scale factor of 0.9 will zoom in
        if np.random.uniform() < self.p_scale_per_sample:
            coords *= scale_factor

        if self.random_crop and self.do_crop:
            for d in range(len(patch_size)):
                crop_center_idx = [
                    np.random.randint(
                        int(patch_size[d] / 2),
                        image.shape[d + 2] - int(patch_size[d] / 2) + 1,
                    )
                ]
                coords[d] += crop_center_idx
        else:
            # Reversing the zero-centering of the coordinates
            for d in range(len(patch_size)):
                coords[d] += image.shape[d + 2] / 2.0 - 0.5

        # Mapping the images to the distorted coordinates
        for b in range(image.shape[0]):
            for c in range(image.shape[1]):
                imageCanvas[b, c] = map_coordinates(
                    image[b, c].astype(float),
                    coords,
                    order=3,
                    mode="constant",
                    cval=cval,
                ).astype(image.dtype)

        data_dict[self.data_key] = imageCanvas
        if data_dict.get(self.label_key) is not None and not skip_label:
            label = data_dict.get(self.label_key)
            labelCanvas = np.zeros(
                (label.shape[0], label.shape[1], *patch_size),
                dtype=np.float32,
            )

            # Mapping the labelmentations to the distorted coordinates
            for b in range(label.shape[0]):
                for c in range(label.shape[1]):
                    labelCanvas[b, c] = map_coordinates(label[b, c], coords, order=0, mode="constant", cval=0.0).astype(
                        label.dtype
                    )
            data_dict[self.label_key] = labelCanvas
        return data_dict

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
			\nShould be (c, x, y, z) or (c, x, y) and is: {data_dict[self.data_key].shape}"

        (
            deform_alpha,
            deform_sigma,
            x_rot_rad,
            y_rot_rad,
            z_rot_rad,
            scale_factor,
        ) = self.get_params(
            deform_alpha=self.deform_alpha,
            deform_sigma=self.deform_sigma,
            x_rot=self.x_rot_in_degrees,
            y_rot=self.y_rot_in_degrees,
            z_rot=self.z_rot_in_degrees,
            scale_factor=self.scale_factor,
        )

        data_dict = self.__CropDeformRotateScale__(
            data_dict,
            self.patch_size,
            deform_alpha,
            deform_sigma,
            x_rot_rad,
            y_rot_rad,
            z_rot_rad,
            scale_factor,
            self.skip_label,
        )
        return data_dict



"""
Takes raw data conforming with Yucca standards and preprocesses according to the generic scheme
"""

import numpy as np
import torch
import os
import logging
import time
from yucca.preprocessing.YuccaPreprocessor import YuccaPreprocessor
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    save_pickle,
    isfile,
)


class ClassificationPreprocessor(YuccaPreprocessor):
    def preprocess_train_subject(self, subject_id):
        subject_id = subject_id.split(os.extsep, 1)[0]
        arraypath = join(self.target_dir, subject_id + ".npy")
        picklepath = join(self.target_dir, subject_id + ".pkl")

        if isfile(arraypath) and isfile(picklepath):
            logging.info(f"Case: {subject_id} already exists. Skipping.")
            return

        start_time = time.time()

        images, label, image_props = self._preprocess_train_subject(subject_id, label_exists=True, preprocess_label=False)

        images = np.array((np.array(images).T, label), dtype="object")
        images[0] = images[0].T

        # save the image
        np.save(arraypath, images)

        # save metadata as .pkl
        save_pickle(image_props, picklepath)

        end_time = time.time()
        logging.info(
            f"Preprocessed case: {subject_id} \n"
            f"size before: {image_props['original_size']} size after: {image_props['new_size']} \n"
            f"spacing before: {image_props['original_spacing']} spacing after: {image_props['new_spacing']} \n"
            f"Saving {subject_id} in {arraypath} \n"
            f"Time elapsed: {round(end_time-start_time, 4)} \n"
        )

    def reverse_preprocessing(self, images: torch.Tensor, image_properties: dict):
        """
        Expected shape of images are:
        (b, c, x)
        """
        image_properties["save_format"] = "txt"
        return images.cpu().numpy(), image_properties



from yucca.planning.YuccaPlanner import YuccaPlanner


class YuccaPlanner_1_1_1(YuccaPlanner):
    def __init__(self, task, preprocessor="YuccaPreprocessor", threads=None, disable_sanity_checks=False, view=None):
        super().__init__(
            task, preprocessor=preprocessor, threads=threads, disable_sanity_checks=disable_sanity_checks, view=view
        )
        self.name = str(self.__class__.__name__) + str(view or "")
        self.view = view
        self.norm_op = "volume_wise_znorm"

    def determine_target_size_from_fixed_size_or_spacing(self):
        self.fixed_target_spacing = [1, 1, 1]
        self.fixed_target_size = None



from yucca.image_processing.transforms.YuccaTransform import YuccaTransform
import numpy as np


class GibbsRinging(YuccaTransform):
    def __init__(self, data_key="image", p_per_sample=1, cutFreq=(96, 129), axes=(0, 3)):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.cutFreq = cutFreq
        self.axes = axes

    @staticmethod
    def get_params(cutFreq, axes):
        cutFreq = np.random.randint(*cutFreq)
        axis = np.random.randint(*axes)
        return cutFreq, axis

    def __gibbsRinging__(self, imageVolume, numSample, axis):
        m = min(0, imageVolume.min())
        imageVolume += abs(m)
        if len(imageVolume.shape) == 3:
            assert axis in [0, 1, 2], "Incorrect or no axis"

            h, w, d = imageVolume.shape
            if axis == 0:
                imageVolume = imageVolume.transpose(0, 2, 1)
                imageVolume = np.fft.fftshift(np.fft.fftn(imageVolume, s=[h, d, w]))
                imageVolume[:, :, 0 : int(np.ceil(w / 2) - np.ceil(numSample / 2))] = 0
                imageVolume[:, :, int(np.ceil(w / 2) + np.ceil(numSample / 2)) : w] = 0
                imageVolume = abs(np.fft.ifftn(np.fft.ifftshift(imageVolume), s=[h, d, w]))
                imageVolume = imageVolume.transpose(0, 2, 1)
            elif axis == 1:
                imageVolume = imageVolume.transpose(1, 2, 0)
                imageVolume = np.fft.fftshift(np.fft.fftn(imageVolume, s=[w, d, h]))
                imageVolume[:, :, 0 : int(np.ceil(h / 2) - np.ceil(numSample / 2))] = 0
                imageVolume[:, :, int(np.ceil(h / 2) + np.ceil(numSample / 2)) : h] = 0
                imageVolume = abs(np.fft.ifftn(np.fft.ifftshift(imageVolume), s=[w, d, h]))
                imageVolume = imageVolume.transpose(2, 0, 1)
            else:
                imageVolume = np.fft.fftshift(np.fft.fftn(imageVolume, s=[h, w, d]))
                imageVolume[:, :, 0 : int(np.ceil(d / 2) - np.ceil(numSample / 2))] = 0
                imageVolume[:, :, int(np.ceil(d / 2) + np.ceil(numSample / 2)) : d] = 0
                imageVolume = abs(np.fft.ifftn(np.fft.ifftshift(imageVolume), s=[h, w, d]))
        elif len(imageVolume.shape) == 2:
            assert axis in [0, 1], "incorrect or no axis"
            h, w = imageVolume.shape
            if axis == 0:
                imageVolume = np.fft.fftshift(np.fft.fftn(imageVolume, s=[h, w]))
                imageVolume[:, 0 : int(np.ceil(w / 2) - np.ceil(numSample / 2))] = 0
                imageVolume[:, int(np.ceil(w / 2) + np.ceil(numSample / 2)) : w] = 0
                imageVolume = abs(np.fft.ifftn(np.fft.ifftshift(imageVolume), s=[h, w]))
            else:
                imageVolume = imageVolume.conj().T
                imageVolume = np.fft.fftshift(np.fft.fftn(imageVolume, s=[w, h]))
                imageVolume[:, 0 : int(np.ceil(h / 2) - np.ceil(numSample / 2))] = 0
                imageVolume[:, int(np.ceil(h / 2) + np.ceil(numSample / 2)) : h] = 0
                imageVolume = abs(np.fft.ifftn(np.fft.ifftshift(imageVolume), s=[w, h]))
                imageVolume = imageVolume.conj().T
        imageVolume -= m
        return imageVolume

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                if np.random.uniform() < self.p_per_sample:
                    cutFreq, axis = self.get_params(self.cutFreq, self.axes)
                    data_dict[self.data_key][b, c] = self.__gibbsRinging__(data_dict[self.data_key][b, c], cutFreq, axis)
        return data_dict



import numpy as np

np.seterr(all="ignore")


def softmax(x, axis=1):
    """Compute softmax values for each sets of scores in x."""
    x_exp = np.exp(x - np.max(x, axis=axis))
    return x_exp / np.sum(x_exp, axis=axis)



from yucca.image_processing.transforms.YuccaTransform import YuccaTransform
import numpy as np
from typing import Tuple


class MotionGhosting(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        p_per_sample=1,
        alpha=(0.85, 0.95),
        numReps=(2, 5),
        axes=(0, 3),
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.alpha = alpha
        self.numReps = numReps
        self.axes = axes

    @staticmethod
    def get_params(alpha: Tuple[float], numReps: Tuple[float], axes: Tuple[float]) -> Tuple[float]:
        alpha = np.random.uniform(*alpha)
        numReps = np.random.randint(*numReps)
        axis = np.random.randint(*axes)
        return alpha, numReps, axis

    def __motionGhosting__(self, imageVolume, alpha, numReps, axis):
        m = min(0, imageVolume.min())
        imageVolume += abs(m)
        if len(imageVolume.shape) == 3:
            assert axis in [0, 1, 2], "Incorrect or no axis"

            h, w, d = imageVolume.shape

            imageVolume = np.fft.fftn(imageVolume, s=[h, w, d])

            if axis == 0:
                imageVolume[0:-1:numReps, :, :] = alpha * imageVolume[0:-1:numReps, :, :]
            elif axis == 1:
                imageVolume[:, 0:-1:numReps, :] = alpha * imageVolume[:, 0:-1:numReps, :]
            else:
                imageVolume[:, :, 0:-1:numReps] = alpha * imageVolume[:, :, 0:-1:numReps]

            imageVolume = abs(np.fft.ifftn(imageVolume, s=[h, w, d]))
        if len(imageVolume.shape) == 2:
            assert axis in [0, 1], "Incorrect or no axis"
            h, w = imageVolume.shape
            imageVolume = np.fft.fftn(imageVolume, s=[h, w])

            if axis == 0:
                imageVolume[0:-1:numReps, :] = alpha * imageVolume[0:-1:numReps, :]
            else:
                imageVolume[:, 0:-1:numReps] = alpha * imageVolume[:, 0:-1:numReps]
            imageVolume = abs(np.fft.ifftn(imageVolume, s=[h, w]))
        imageVolume -= m
        return imageVolume

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                if np.random.uniform() < self.p_per_sample:
                    alpha, numReps, axis = self.get_params(self.alpha, self.numReps, self.axes)
                    data_dict[self.data_key][b, c] = self.__motionGhosting__(
                        data_dict[self.data_key][b, c], alpha, numReps, axis
                    )
        return data_dict



from yucca.image_processing.transforms.YuccaTransform import YuccaTransform
import numpy as np
from typing import Tuple
from scipy.ndimage import gaussian_filter


class Blur(YuccaTransform):
    def __init__(self, data_key="image", p_per_sample=1, p_per_channel=0.5, sigma=(0.5, 1.0)):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.p_per_channel = p_per_channel
        self.sigma = sigma

    @staticmethod
    def get_params(sigma: Tuple[float]):
        sigma = np.random.uniform(*sigma)
        return sigma

    def __blur__(self, imageVolume, sigma):
        for c in range(imageVolume.shape[0]):
            if np.random.uniform() < self.p_per_channel:
                imageVolume[c] = gaussian_filter(imageVolume[c], sigma, order=0)
        return imageVolume

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            if np.random.uniform() < self.p_per_sample:
                sigma = self.get_params(self.sigma)
                data_dict[self.data_key][b] = self.__blur__(data_dict[self.data_key][b], sigma)
        return data_dict
